function [GRD0, GRD1] = blockgridsmask(GRD, msk)

% blockgridsmask -- Select a subset of blocks in a grid set
%
% Usage:
%       [GRD0, GRD1] = blockgridsmask(GRD, msk)
%
% Input:
%       GRD          The structure returned by the blockgrids function
%       msk          A logical mask image with missing regions
%                    denoted by 0 values
%
% Output:
%       GRD0         A copy of the GRD input, but with every block
%                    not containing any missing pixels in msk removed
%       GRD1         A copy of the GRD input, but with every block
%                    containing any missing pixels in msk removed
% 
% Blocks are removed from a grid by truncation of the grid where
% possible (where an entire row or column is to be removed), or
% otherwise by marking the corresponding block position in the grid
% as zero.

% Determine number of grids in grid set
Ng = size(GRD.Grids,1);
% Create initial output grid sets
GRD0 = GRD;
GRD1 = GRD;

Gs0 = zeros(Ng,1);
Gs1 = zeros(Ng,1);

% Iterate over all grids in input grid set
for g = 1:Ng,
  % Current grid row coords
  Gr = GRD.Grids{g,1};
  % Current grid column coords
  Gc = GRD.Grids{g,2}; 
  % Iterate over all block locations in current grid
  for nr = 1:size(Gr,1),
    for nc = 1:size(Gc,2),
      % Current block coords in current grid
      r = Gr(nr,nc);
      c = Gc(nr,nc);
      % Check whether current block has no pixels missing in mask
      mb = msk(r:(r+GRD.Nbr-1),c:(c+GRD.Nbc-1));
      if min(mb(:)) == 1,
	% Mark block as deleted from current grid in primary grid set
	GRD0.Grids{g,1}(nr,nc) = 0;
	GRD0.Grids{g,2}(nr,nc) = 0;
	% Update block count image
	GRD0.BlkCnt(r:(r+GRD.Nbr-1),c:(c+GRD.Nbc-1)) = ...
	    GRD0.BlkCnt(r:(r+GRD.Nbr-1),c:(c+GRD.Nbc-1)) - 1;
      else
	% Mark block as deleted from current grid in secondary grid set
	GRD1.Grids{g,1}(nr,nc) = 0;
	GRD1.Grids{g,2}(nr,nc) = 0;
	% Update block count image
	GRD1.BlkCnt(r:(r+GRD.Nbr-1),c:(c+GRD.Nbc-1)) = ...
	    GRD1.BlkCnt(r:(r+GRD.Nbr-1),c:(c+GRD.Nbc-1)) - 1;
      end
    end
  end
  
  % Truncate current GRD0 grid where possible
  zc = sum(GRD0.Grids{g,1},1) == 0;
  zr = sum(GRD0.Grids{g,1},2) == 0;
  GRD0.Grids{g,1} = GRD0.Grids{g,1}(~zr, ~zc);
  GRD0.Grids{g,2} = GRD0.Grids{g,2}(~zr, ~zc);
  
  Gs0(g) = prod(size(GRD0.Grids{g,1}));
  
  % Truncate current GRD1 grid where possible
  zc = sum(GRD1.Grids{g,1},1) == 0;
  zr = sum(GRD1.Grids{g,1},2) == 0;
  GRD1.Grids{g,1} = GRD1.Grids{g,1}(~zr, ~zc);
  GRD1.Grids{g,2} = GRD1.Grids{g,2}(~zr, ~zc);

  Gs1(g) = prod(size(GRD1.Grids{g,1}));
  
end

% Delete empty grids
GRD0.Grids = {GRD0.Grids{Gs0>0,1}; GRD0.Grids{Gs0>0,2}}';
GRD1.Grids = {GRD1.Grids{Gs1>0,1}; GRD1.Grids{Gs1>0,2}}';

return
