function [ipt, ptm, opt, rgngrd] = inpaintgrey(img, msk, opt)

% Start timer
tic;  

% Set parameters
if nargin < 3,
  opt = [];
end
opt = defaultopts(opt);
Nbr = opt.BlockSize(1);
Nbc = opt.BlockSize(2);
icgt = opt.InitCGTol;
bcgt = opt.BPDNCGTol;
lambda = opt.BPDNLambda;
bmi = opt.BPDNMaxIter;
nbf = opt.NormalBlocks;
nmn = opt.NNMaxNum;
nmd = opt.NNMaxAngDist;
nsr = opt.NNSearchRadius;
mkf = opt.MinKnownFraction;
vrb = opt.Verbose;
 

% Construct grid sets
GRD = blockgrids(img, Nbr, Nbc, opt.GridStep);
[GRD0, GRD1] = blockgridsmask(GRD, msk);

% Construct dictionary source block set
BS = blksetinit([Nbr Nbc], nbf);
BS = blksetadd(BS, img, GRD1);
BS = blksetend(BS);

% Construct inpainting structure
nnfn = @(x, y, w) nnangular(x, y, nmn, nmd, w, 1);
INPT = initinpnt(GRD0, img, msk, BS, nnfn, mkf, nsr);
INPT.gamma0 = opt.FuncGamma(1);
INPT.gamma1 = opt.FuncGamma(2);

% Construct initial solution via CG
Ng = size(INPT.Phi,1);
b = [(INPT.gamma0/INPT.Nv1)*inpntRB(img, INPT); zeros((Ng-1)*INPT.Nvg,1)];
ATb = inpntAT(b, INPT);
ATA = @(x) inpntAT(inpntA(x, INPT), INPT);
[alpha0, cgf] = cgs(ATA, ATb, icgt, 20000);

% Solve using IRLS
p = 1; epsr = 1e-8;
A = @(x) inpntA(x, INPT);
AT = @(x) inpntAT(x, INPT);
opt = struct('Verbose', vrb, 'RegTrmNorm', p, 'RegTrmWThresh', epsr, ...
	     'CGTol', bcgt, 'CGTolScale', 2, 'FracStopTol', 0.005, ...
	     'MaxMainIter', bmi);
[alpha,its] = irlsbpdn({A,AT}, b, lambda, opt, alpha0);


% Reconstruct image
[rgnipt, rgnbnm, rgngrd] = inpntimg(alpha, INPT);
ipt = img;
ipt(~msk) = rgnipt(~msk);

ptm = toc;

return


function opt = defaultopts(opt)
  
  if ~isfield(opt,'BlockSize'),
    opt.BlockSize = [8 8];
  end
  if ~isfield(opt,'GridStep'),
    opt.GridStep = [1 floor(min(opt.BlockSize)/2)];
  end
  if ~isfield(opt,'FuncGamma'),
    opt.FuncGamma = [1.0 1.0];
  end
  if ~isfield(opt,'InitCGTol'),
    opt.InitCGTol = 1e-3;
  end
  if ~isfield(opt,'BPDNCGTol'),
    opt.BPDNCGTol = 1e-6;
  end
  if ~isfield(opt,'BPDNLambda'),
    opt.BPDNLambda = 1e-5;
  end
  if ~isfield(opt,'BPDNMaxIter'),
    opt.BPDNMaxIter = 3;
  end
  if ~isfield(opt,'NormalBlocks'),
    opt.NormalBlocks = 1;
  end
  if ~isfield(opt,'NNMaxNum'),
    opt.NNMaxNum = 8*prod(opt.BlockSize);
  end
  if ~isfield(opt,'NNMaxAngDist'),
    opt.NNMaxAngDist = 0.5;
  end
  if ~isfield(opt,'NNSearchRadius'),
    opt.NNSearchRadius = 15;
  end
  if ~isfield(opt,'MinKnownFraction'),
    opt.MinKnownFraction = 0.75;
  end
  if ~isfield(opt,'Verbose'),
    opt.Verbose = 0;
  end

return


function u = inpntA(alpha, INPT)

% inpntA -- Apply inpainting operator A to compound coefficient vector
%
% Usage:
%       u = inpntA(alpha, INPT)
%
% Input:
%       alpha        Compound coefficient vector
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       u            Compound vector resulting from operator A

  
Ng = size(INPT.Phi,1); % Number of grids
u1 = zeros(INPT.Nv1,1); % Known pixel mismatch vector
u2 = zeros((Ng-1)*INPT.Nvg,1); % Grid overlap mismatch vector

% Initialise cell array of sum of QT applied to blk recons for each grid
QTUB = cell(Ng,1);

% Iterate over all grids
for g = 1:Ng,
  Nb = size(INPT.Crd{g},1); % Number of blocks in current grid
  qtub = zeros(INPT.Nvg,1); % Initialise current grid inpt rgn vector
  % Iterate over all blocks in current grid
  for k = 1:Nb,
    ab = alpha(inpntcx(g, k, INPT)); % Dict coefs for current block    
    ub = INPT.Phi{g}{k}*ab; % Current block reconstruction
    % Set current block part of u1 from R applied to current blk recon
    u1(inpntrx(g, k, INPT)) =  inpntR(ub, g, k, INPT);
    % Add contrib from current block to sum of QT applied to blk recons
    qtub = qtub + inpntQT(ub, g, k, INPT);
  end
  QTUB{g} = qtub; % Set QT of blk recons sum for current grid
end

% Apply G matrix to u2
for g = 2:Ng,
  u2(inpntgx(g-1, INPT)) = INPT.Grm{g-1}.*QTUB{g} - INPT.Grm{g}.*QTUB{g-1};
end

% Construct composite u from concatenation of u1 and u2
u = [(INPT.gamma0/length(u1))*u1; (INPT.gamma1/length(u2))*u2];

return


function alpha = inpntAT(u, INPT)

% inpntAT -- Apply inpainting operator A^T to compound vector
%
% Usage:
%       alpha = inpntAT(u, INPT)
%
% Input:
%       u            Compound vector resulting from operator A
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       alpha        Compound coefficient vector
  
  
Ng = size(INPT.Phi,1); % Number of grids
u1 = u(1:INPT.Nv1); % Known pixel mismatch vector
u2 = u((INPT.Nv1+1):end); % Grid overlap mismatch vector

% Initialise cell array of GT applied to each grid pair part of u2
GTU1 = cell(Ng,1);

% Compute GT applied to each grid pair in u2
GTU1{1} = -INPT.Grm{1}.*u2(inpntgx(1, INPT));
for g = 2:(Ng-1),
  GTU1{g} = INPT.Grm{g}.*(u2(inpntgx(g-1, INPT)) - u2(inpntgx(g, INPT)));
end
GTU1{Ng} = INPT.Grm{Ng}.*u2(inpntgx(Ng-1, INPT));

alpha = zeros(INPT.Nv0,1); % Initialise concatenated coef vector
% Iterate over all grids
for g = 1:Ng,
  Nb = size(INPT.Crd{g},1); % Number of blocks in current grid
  % Iterate over all blocks in current grid
  for k = 1:Nb,
    % Apply R^T to current block in u1
    rtu1b = inpntRT(u1(inpntrx(g, k, INPT)), g, k, INPT);
    % Apply Q to current block in GTU1
    qgtu2b = inpntQ(GTU1{g}, g, k, INPT);
    % Sum known pixel mismatch and grid overlap mismatch
    % contributions for current block
    ub = (INPT.gamma0/length(u1))*rtu1b + (INPT.gamma1/length(u2))* qgtu2b;
    % Compute coefficients for current block by applying Phi^T to ub 
    alpha(inpntcx(g, k, INPT)) = (INPT.Phi{g}{k})'*ub;
  end
end
    
return


function v = inpntQ(u, k, l, INPT)

% inpntQ -- Apply inpainting operator Q to inpainting region vector 
%
% Usage:
%       v = inpntQ(u, k, l, INPT)
%
% Input:
%       u            Inpainting region vector
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       v            Block k,l, non-zero where intersecting with
%                    inpainting region 

xd = INPT.Map{k}{l}(:,2);
xr = INPT.Map{k}{l}(:,1);
v = zeros(size(INPT.Phi{k}{l},1),1);
v(xr) = u(xd);

return


function v = inpntQT(u, k, l, INPT)

% inpntQ -- Apply inpainting operator Q to image block 
%
% Usage:
%       v = inpntQT(u, k, l, INPT)
%
% Input:
%       v            Image block k,l
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       v            Inpainting region vector contribution from
%                    block k,l
    
xd = INPT.Map{k}{l}(:,1);
xr = INPT.Map{k}{l}(:,2);
v = zeros(INPT.Nvg,1);
v(xr) = u(xd);

return


function v = inpntR(u, k, l, INPT)

% inpntR -- Apply inpainting operator R to image block k,l
%
% Usage:
%       v = inpntR(u, k, l, INPT)
%
% Input:
%       u            Image block k,l
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       v            Known part of block k,l

rg = INPT.Rgn{k}{l};
v = u(rg(:) > 0);

return


function v = inpntRT(u, k, l, INPT)

% inpntRT -- Apply inpainting operator R^T to known part of image block k,l
%
% Usage:
%       v = inpntRT(u, k, l, INPT)
%
% Input:
%       u            Known part of block k,l
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       v            Image block k,l


rg = INPT.Rgn{k}{l};
v = zeros(size(INPT.Phi{k}{l},1),1);
v(rg(:) > 0) = u;

return


function v = inpntRB(u, INPT)

% inpntRB -- Apply product of inpainting operators R and B to an image
%
% Usage:
%       v = inpntRB(u, INPT)
%
% Input:
%       u            Input image
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       u            Compound vector resulting from operators R and B

  
v = zeros(INPT.Nv1,1);
Ng = size(INPT.Phi,1);
for g = 1:Ng,
  Nb = size(INPT.Crd{g},1);
  for k = 1:Nb,
    rgn = INPT.Rgn{g}{k};
    r1 = INPT.Crd{g}(k,1);
    c1 = INPT.Crd{g}(k,2);
    r2 = INPT.Crd{g}(k,3);
    c2 = INPT.Crd{g}(k,4);
    ub = u(r1:r2,c1:c2);    
    vb = ub(rgn > 0);    
    v(INPT.Ix1{g}(k,1):INPT.Ix1{g}(k,2)) = vb;
  end
end

return


function idx = inpntrx(k, l, INPT)

% inpntrx -- Find start index of block k,l in concatenated
%            known-region vector 
%
% Usage:
%       idx = inpntrx(k, l, INPT)
%
% Input:
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       idx          Index in concatenated known-region vector

  
idx = INPT.Ix1{k}(l,1):INPT.Ix1{k}(l,2);

return


function idx = inpntcx(k, l, INPT)
  
% inpntcx -- Find start index of block k,l in concatenated coefficient vector
%
% Usage:
%       idx = inpntcx(k, l, INPT)
%
% Input:
%       k            Grid index
%       l            Index of block in grid k
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       idx          Index in concatenated coefficient vector
 

idx = INPT.Ix0{k}(l,1):INPT.Ix0{k}(l,2);

return



function idx = inpntgx(k, INPT)

% inpntgx -- Find start index of entry k in inpainting region grid
%            difference vector 
%
% Usage:
%       idx = inpntgx(k, INPT)
%
% Input:
%       k            Grid difference entry index
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       idx          Index in inpainting region difference vector 
 

idx = ((k-1)*INPT.Nvg+1):(k*INPT.Nvg);

return


function [img, bnm, grd] = inpntimg(alpha, INPT)

% inpntimg -- Construct final inpainted image from a compound coefficient
%             vector
%
% Usage:
%       [img, bnm, grd] = inpntimg(alpha, INPT)
%
% Input:
%       alpha        Compound coefficient vector
%       INPT         Inpainting information structure (see initinpnt)
%
% Output:
%       img          Inpainted image
%       bnm          Image providing count of number of grids
%                    contributing to each pixel
%       grd          Cell array of individual grid reconstructions

Ng = size(INPT.Phi,1); % Number of grids
img = zeros(size(INPT.Gms{1}));
bnm = zeros(size(INPT.Gms{1}));
grd = cell(Ng,1);
  
% Iterate over all grids
for g = 1:Ng,
  grd{g} = zeros(size(INPT.Gms{1}));
  Nb = size(INPT.Crd{g},1); % Number of blocks in current grid
  % Iterate over all blocks in current grid
  for k = 1:Nb,

    ub = INPT.Phi{g}{k}*alpha(inpntcx(g, k, INPT));
    r1 = INPT.Crd{g}(k,1);
    c1 = INPT.Crd{g}(k,2);
    r2 = INPT.Crd{g}(k,3);
    c2 = INPT.Crd{g}(k,4);    
    grd{g}(r1:r2,c1:c2) = reshape(ub,r2-r1+1,c2-c1+1);
    bnm(r1:r2,c1:c2) = bnm(r1:r2,c1:c2) + 1;
  
  end

  img = img + grd{g};
    
end

img = img ./ bnm;

return


function INPT = initinpnt(GRD, img, msk, BS, nnfn, kft, scr)

% initinpnt -- Initialise a structure containting the information
%              necessary for exemplar-based inpainting
%
% Usage:
%       INPT = initinpnt(GRD, img, msk, BS, nnfn, scr)
%
% Input:
%       GRD         Grid cell array (see blockgrids function)
%       img         Image to be inpainted
%       msk         Region mask image
%       BS          Block set struct (see blksetinit function)
%       nnfn        Function handle for nearest neighbour function
%       kft         Threshold for fraction of block required to be
%                   known when constructing dictionary via
%                   nearest neighbour search
%       scr         Spatial radius for dictionary construction by
%                   spiral scan of spatial neigbourhood (for blocks
%                   below kft threshold)  
%
% Output:
%       INPT        A struct representing the inpainting information


if nargin < 7,
  scr = 12;
end
if nargin < 6,
  kft = 0.75;
end


% Initialise variables data structures
Ng = size(GRD.Grids,1); % Number of grids
Nbr = GRD.Nbr;          % Number of rows in block
Nbc = GRD.Nbc;          % Number of cols in block
CRgn = cell(Ng,1);      % Known region mask for each block
CPhi = cell(Ng,1);      % Dictionary for each block
CCrd = cell(Ng,1);      % Image coordinates of each block
CGms = cell(Ng,1);      % Inpainting region mask for each grid
CIx0 = cell(Ng,1);      % Coordinates for each block in
                        % concatenated coefficient vector
CIx1 = cell(Ng,1);      % Coordinates for each block in
                        % concatenated known-region vector
CMap = cell(Ng,1);      % Map from inpainting region vector for
                        % each block
CGrm = cell(Ng,1);      % Inpainting region vector mask for each grid

% Initialise vector coordinate counters
i0 = 1;
i1 = 1;

% Iterate over all grids
for g = 1:Ng,

  % Initialise per-grid data structures
  Nb = sum(sum(GRD.Grids{g,1} > 0));
  CRgn{g} = cell(Nb,1);
  CPhi{g} = cell(Nb,1);
  CCrd{g} = zeros(Nb,4);
  CGms{g} = zeros(size(msk));
  CIxB{g} = zeros(Nb,2);
  CIxC{g} = zeros(Nb,2);
  CMap{g} = cell(Nb,1);
  
  k = 1; % Initialise block counter for current grid
  
  % Iterate over all blocks in current grid
  for l = 1:size(GRD.Grids{g,1},1),
    for m = 1:size(GRD.Grids{g,1},2),

      % Extract top-left coordinates of current block
      r1 = GRD.Grids{g,1}(l,m);
      c1 = GRD.Grids{g,2}(l,m);

      % Process current block if not marked as removed from grid
      if r1 ~= 0 && c1 ~= 0,

	% Construct bottom-right coordinates of current block
	r2 = r1 + Nbr - 1;
	c2 = c1 + Nbc - 1;
        CRgn{g}{k} = msk(r1:r2,c1:c2); % Current block known region mask

        % Compute fraction of known pixels in current block
        kpf = (sum(vec(CRgn{g}{k}) == 1)/prod(size(CRgn{g}{k})));
        
        % Dictionary for current block depends on fraction of known pixels
	if kpf >= kft,
          % If sufficient fraction of pixels in block are known,
          % construct dictionary via nearest neighbour search in
          % provided block set
	  u = img(r1:r2,c1:c2);
	  w = msk(r1:r2,c1:c2);
          B = nnfn(BS.BlkMx, u(:), w(:));
          if size(B,2) == 0,
              CPhi{g}{k} = nbhdblocks(img, msk, r1, c1, r2, c2, BS, ...
                                      Nbr, Nbc, scr);
          else
              CPhi{g}{k} = B;
          end
	else
          % If insufficient fraction of pixels in block are known,
          % construct a dictionary by scanning the spatial
          % neighbourhood of the current block in a spiral path,
          % adding blocks without any unknown pixels
          B = nbhdblocks(img, msk, r1, c1, r2, c2, BS, Nbr, Nbc, scr);
	  CPhi{g}{k} = B;
	end
				       
	CCrd{g}(k,:) = [r1 c1 r2 c2]; % Current block coordinates
	CGms{g}(r1:r2,c1:c2) = 1; % Mark current block in grid mask
 
	% Construct inpainting vector to block map for current block
        bx = zeros(size(msk));
        bx(r1:r2,c1:c2) = reshape(1:((r2-r1+1)*(c2-c1+1)),r2-r1+1,c2-c1+1);
        CMap{g}{k} = [bx((bx>0).*(1-msk) > 0) find(bx(msk == 0))];

	% Compute dictionary coefficient vector size for current block
        n0 = size(CPhi{g}{k},2);
	% Compute known region vector size for current block
        n1 = sum(CRgn{g}{k}(:) == 1);

	% Construct vector coordinate records
	CIx0{g}(k,:) = [i0 i0+n0-1];
	CIx1{g}(k,:) = [i1 i1+n1-1];

	% Increment vector coordinate counters
	i0 = i0 + n0;
	i1 = i1 + n1;

	% Increment current block counter
	k = k+1;

      end

    end
  end

  % Compute inpainting region vector mask for current grid
  CGrm{g} = CGms{g}(msk == 0);

end

Nv0 = CIx0{Ng}(end,2); % Concatenated dict coefficient vector length
Nv1 = CIx1{Ng}(end,2); % Concatenated known region vector length
Nvg = size(CGrm{g},1); % Inpainting region vector length

% Construct output structure
INPT = struct('Rgn', {CRgn}, 'Phi', {CPhi}, 'Crd', {CCrd}, 'Gms', {CGms}, ...
              'Ix0', {CIx0}, 'Nv0', Nv0, 'Ix1', {CIx1}, 'Nv1', Nv1, ...
              'Map', {CMap}, 'Grm', {CGrm}, 'Nvg', Nvg);

return


function [XSv, YSv] = spiralpath(Sr)

SM = spiral(2*Sr+1);
[XS, YS] = meshgrid(-Sr:Sr,-Sr:Sr);
[Sv, Si] = sort(SM(:));
XSv = XS(Si);
YSv = YS(Si);

return


function B = nbhdblocks(img, msk, r1, c1, r2, c2, BS, Nbr, Nbc, scr)

[XSv, YSv] = spiralpath(scr);
B = [];
for nb = 1:length(XSv),
    rs1 = r1 + YSv(nb);
    cs1 = c1 + XSv(nb);
    rs2 = rs1 + Nbr - 1;
    cs2 = cs1 + Nbc - 1;
    if rs1>=1 && cs1>=1 && rs2 <= size(img,1) && cs2 <= size(img,2),
        blkm = msk(rs1:rs2,cs1:cs2);
        if sum(blkm(:)) == prod(size(blkm)),		
            B = [B vec(img(rs1:rs2,cs1:cs2))];
            if BS.nflg,
                B(:,end) = B(:,end) / norm(B(:,end));
            end
        end
    end
end

%B = B(:,std(B,0,1) > 0.1);

return
