function [U,D] = blksrcmatch(BS, u, m, d, w, dfn)

% blksrcmatch -- Construct a nearest neighbour dictionary for the
%                specified image block
%
% Usage:
%       [U,D] = blksrcmatch(BS, u, m, d, w, dfn)
%
% Input:
%       BS          Block source struct
%       u           The reference vector for which neighbours are desired
%       m           Maximum number of neighbours
%       d           Neighbourhood radius (angular distance)
%       w           Weighting vector
%       dfn         A string selecting the measure of match. Valid
%                   values are 'ang' (angular distance), 'absang'
%                   (angular distance with an absolute value on
%                   inner product so that angles > pi/2 are closer
%                   than pi/2) and 'euclid' (Euclidean distance).
%
% Output:
%       U           A matrix containing the neighbour vectors
%       D           A vector giving distances to the neighbour vectors

if nargin < 6,
  dfn = 'ang'; 
end
if nargin < 5 || isempty(w),
  w = ones(size(u,1),size(u,2));
end
if nargin < 4,
  d = [];
end
if nargin < 3,
  m = [];
end

if strcmp(dfn,'ang'),
  dflg = 0;
  aflg = 1;
elseif strcmp(dfn,'absang'),
  dflg = 1;
  aflg = 1;
else
  dflg = 0;
  aflg = 0;
end

% Compute composite weight matrix combining base block active
% region mask (specified in blksrcadd) and weighting matrix for
% current reference block
Wbx = repmat(BS.Wb,[1 1 size(u,3)]);
wx = repmat(w,[1 1 size(u,3)]);
Wp = wx.*Wbx;


mflg = 0;
Mc = cell(BS.in,1);
% Search through all recorded block source images
for k = 1:BS.in,
  % If current block source image DFTs have not been cached, call
  % blkmatch with the image and cache the returned DFT data,
  % otherwise call blkmatch with the cached DFT data.
  if isempty(BS.U12C{k}),
    [M, U12] = blkmatch(BS.ImgC{k}, u, Wp, aflg);
    BS.U12C{k} = U12;
    mflg = 1;
  else
    M = blkmatch(BS.U12C{k}, u, Wp, aflg);
  end
  % Transform match values into distances
  if aflg,
    if dflg,
        Da = (1 - abs(M));
    else
        Da = ((1 - M)/2);
    end
  else
    Da = M; 
  end
  % Mark invalid match positions with infinite distance
  Da(~BS.VldC{k}) = Inf;
  % Record distance map for current block source image
  Mc{k} = Da;
end

% If new DFT data has been cached, write this data into BS in the
% calling scope
if mflg,
  assignin('caller',inputname(1),BS)
end

% Use distance maps to assemble dictionary of nearest neighbour
% blocks
[U, D] = mapblkdict(BS.ImgC, Wbx, Mc, BS.Gr, BS.Gc, m, d);

return
