function [U, D, C] = mapblkdict(CI, Wbx, CD, CGr, CGc, m, d)

% mapblkdict -- Assemble blocks indicated by distance maps into a
%               nearest neighbour dictionary
%
% Usage:
%       [U, D] = mapblkdict(CI, Wbx, CD, CGr, CGc, m, d)
%
% Input:
%       CI          A cell array of block source images
%       Wbx         Block weight matrix
%       CD          A cell array of distance maps
%       CGr         A cell array of image row coordinates
%       CGc         A cell array of image column coordinates
%       m           Maximum number of neighbours
%       d           Neighbourhood radius (angular distance)
%
% Output:
%       U           A matrix containing the neighbour vectors
%       D           A vector giving distances to the neighbour vectors
%       C           A matrix giving source image index and block
%                   coordinates for each block assembled into
%                   neigbour vector matrix U

if nargin < 7,
  d = [];
end
if nargin < 6,
  m = [];
end

Dv = [];
if ~isempty(d) || ~isempty(m),
  for k = 1:numel(CD),
    % If maximum distance specified, set all distances above the
    % threshold to infinity
    if ~isempty(d),
      CD{k}(CD{k} > d) = Inf;
    end
    % If limit on number of neighbours is specified, add list of current
    % distances to a global list
    if ~isempty(m),
      vcd = vec(CD{k});
      Dv = [Dv; vcd(isfinite(vcd))];
    end
  end
end

% Determine threshold distance (if any) for returned matches
if ~isempty(m),
  % If limit on number of matches is specified, determine appropriate
  % threshold from list of distances for all block source images
  Ds = sort(Dv);
  if m < length(Ds),
    dn = (Ds(m)+Ds(m+1))/2;
  else
    if length(Dv) > 0,
      dn = Ds(end)+1;
    else
      dn = Inf; % This case should be handled more efficiently 
    end
  end
  % If maximum distance is also specified, the final distance
  % threshold is the minimum of this maximum and the threshold
  % computed based on the maximum number of required matches.
  if isempty(d),
    dm = dn;
  else
    dm = min(dn,d);
  end
else
  if ~isempty(d),
    dm = d;
  else
    dm = Inf;
  end
end

% Assemble matching blocks in each block source image into the
% output matrix of vectorised blocks (each vector is a column)
U = [];
D = [];
C = [];
for k = 1:numel(CD),
  R = CD{k} < dm;
  Ck = [CD{k}(R) CGr{k}(R) CGc{k}(R)];
  D = [D; Ck(:,1)];
  C = [C; [k*ones(size(Ck,1),1) Ck(:,2:end)]];
  Uk = zeros(sum(Wbx(:)==1), size(Ck,1));
  for l = 1:size(Ck,1),
    r1 = Ck(l,2); c1 = Ck(l,3);
    r2 = r1 + size(Wbx,1) - 1; c2 = c1 + size(Wbx,2) - 1;
    b = CI{k}(r1:r2,c1:c2,:);
    Uk(:,l) = b(Wbx == 1);
  end
  U = [U Uk];
end

% Sort distances and corresponding block order
[D, Idx] = sort(D);
if length(Idx) > m,
  Idx = Idx(1:m);
  D = D(1:m);
end
U = U(:,Idx);

if nargout > 2,
  C = C(Idx,:);
end

return
