function [U, D, Idx] = nnangular(V, u, m, d, w, dflg)

% nnangular -- Find a set of neighbouring vectors according to
%              angular distance
%
% Usage:
%       [U, D, Idx] = nnangular(V, u, m, d, w, dflg)
%
% Input:
%       V           Vectors to search arranged as columns of a matrix
%       u           The reference vector for which neighbours are desired
%       m           Maximum number of neighbours
%       d           Neighbourhood radius (angular distance)
%       w           Weighting vector
%       dflg        If true, distance measure includes absolute
%                   value on inner product so that angles > pi/2
%                   are closer than an pi/2
%
% Output:
%       U           A matrix containing the neighbour vectors
%       D           A vector giving distances to the neighbour vectors
%       Idx         A vector giving indices in block set of found vectors

[Nd,Nv] = size(V);

if nargin < 6,
  dflg = 0;
end
if nargin < 5 || isempty(w),
  w = ones(size(u));
end
if nargin < 4,
  d = [];
end
if nargin < 3,
  m = [];
end

if size(u,1) < size(u,2),
  u = u';
end
if size(w,1) < size(w,2),
  w = w';
end
if m > Nv,
  m = Nv;
end

IP = (V'*(w.*u))./(sqrt(sum((V'*diag(w)).^2,2))*sqrt(sum((w.*u).^2)));
if dflg,
  Da = 1 - abs(IP);
else
  Da = (1 - IP)/2;
end
[D, Idx] = sort(Da);

if ~isempty(m),
  Idx = Idx(1:m);
  D = D(1:m);
end

if ~isempty(d),
  Idx = Idx(D < d);
  D = D(D < d);
end

U = V(:,Idx);

return
