function [U, itstat] = irntv(S, KC, lambda, p, q, epsf, epsr, nit, U0, rrs)

% irntv -- Compute the minimum of a generalised TV functional using
%          the IRN algorithm (for details see P. Rodríguez and
%          B. Wohlberg, "Efficient Minimization Method for a
%          Generalized Total Variation Functional", IEEE Transactions
%          on Image Processing, vol. 18, no. 2, pp. 322-332, Feb 2009,
%          doi:10.1109/TIP.2008.2008420). This code is also
%          available in the Numipad distribution at numipad.sourceforge.net.
%    
% Usage:
%       [U, itstat] = irntv(S, KC, lambda, p, q, epsf, epsr, nit, U0, rrs)
%
% Input:
%       S           Input image
%       KC          Cell array {K, KT} of function handles for
%                   forward linear operator and its
%                   transpose. These operators act on images rather
%                   than vectors, but the transpose should be
%                   considered to be in the sense of the operator
%                   on a vectorised image. Set to [] for a
%                   denoising problem.
%       lambda      Regularisation term weighting factor
%       p           Data fidelity term norm
%       q           Regularisation term norm
%       epsf        Data fidelity epsilon
%       epsr        Regularisation epsilon
%       nit         Number of iterations
%       U0          Initial solution
%       rrs         Relative residual scaling after each iteration
%
% Output:
%       U           Output image
%       itstat      Iteration status

if nargin < 10,
  rrs = 5;
  if nargin < 9,
    U0 = [];
    if nargin < 8,
      nit = 5;
      if nargin < 7,
	epsr = 1e-3;
	if nargin < 6,
	  epsf = 1e-2;
	  if nargin < 5,
	    q = 1;
	    if nargin < 4,
	      p = 1;
	      if nargin < 3,
		lambda = 1;
		if nargin < 2,
		  K = [];
		end
	      end
	    end
	  end
	end
      end
    end
  end
end

dflag = 1;
if nargout > 1,
  dflag = 0;
end

%pcgtol = 1e-5;
pcgitn = 5000;

% Check argument KC
if ~isempty(KC),
  if ~(iscell(KC) && isa(KC{1}, 'function_handle') && isa(KC{2}, 'function_handle')),
   warning('argument KC must be empty or a cell array {K,KT} of function handles');
   U = [];
   return;
 end
end

s = S(:);

% Initialise iteration status vector
itstat = [];

% Set up status display
iterdispopen(dflag);

% Start clock
tic;

if isempty(KC),  % Denoising problem

  sbstflg = 1; % Flag determining use of substitution (indirect) method

  [Nr Nc] = size(S);
  
  if isempty(U0),
    % Construct initial solution
    [u, flg, rlr, pit] = pcg(@(x) IDTD(x,Nr,Nc,lambda),s,1e-3,500,[],[],s);
    U = reshape(u,Nr,Nc);
  else
    U = U0;
    u = U(:);
  end

  % Iterate
  for k = 1:nit,

    % Set weight matrices
    WF = (2/p)*fF(U-S, p, epsf);
    WR = (2/q)*fR(Dx(U).^2 + Dy(U).^2, q, epsr);

    % Compute relative residual of current solution
    if sbstflg, % Use substitution (indirect) method
      wfs = vec(WF.^(1/2)).*s;
      v = vec(WF.^(1/2)).*u;
      relres = pcgrelres(@(x) IWDTWDW(x,Nr,Nc,WF,WR,lambda), wfs, v);
    else        % Use standard (direct) method
      wfs = vec(WF).*s;
      relres = pcgrelres(@(x) IDTWD(x,Nr,Nc,WF,WR,lambda), wfs, u);
    end
    pcgtol = relres/rrs;

    % Compute and display functional values
    iterdisp(S, [], lambda, p, q, U, WF, WR, k, relres, [], [], sbstflg, dflag);

    % Update current solution
    if sbstflg, % Use substitution (indirect) method
      [v, flg, rlr, pit] = pcg(@(x) IWDTWDW(x,Nr,Nc,WF,WR,lambda),wfs,pcgtol,pcgitn,[],[],v);
      V = reshape(v,Nr,Nc);
      u = vec(WF.^(-1/2)).*v;
    else        % Use standard (direct) method
      [u, flg, rlr, pit] = pcg(@(x) IDTWD(x,Nr,Nc,WF,WR,lambda),wfs,pcgtol,pcgitn,[],[],u);
    end

    U = reshape(u,Nr,Nc);

    % Compute and display functional values and CG status
    is = iterdisp(S, [], lambda, p, q, U, WF, WR, [], relres, flg, pit, sbstflg, dflag);
    itstat = [itstat; [k is]];

  end

else % General inverse problem with linear operator K

  % KC is cell array of K and K^T function handles
  K = KC{1};
  KT = KC{2};
 
  [Nr Nc] = size(KT(S));
  
  if isempty(U0),
    % Construct initial solution
    kts = vec(KT(S));
    [u, flg, rlr, pit] = pcg(@(x) KTKDTD(x,Nr,Nc,K,KT,lambda),kts,1e-3,500);
    U = reshape(u,Nr,Nc);
  else
    U = U0;
    u = U(:);
  end

  % Iterate
  for k = 1:nit,

    % Set weight matrices
    WF = (2/p)*fF(K(U)-S, p, epsf);
    WR = (2/q)*fR(Dx(U).^2 + Dy(U).^2, q, epsr);

    % Compute relative residual of current solution
    wfs = vec(KT(WF.*S));
    relres = pcgrelres(@(x) KTWKDTWD(x, Nr, Nc, K, KT, WF, WR, lambda), wfs, u);
    pcgtol = relres/rrs;

    % Compute and display functional values
    iterdisp(S, K, lambda, p, q, U, WF, WR, k, relres, [], [], [], dflag);

    % Update current solution
    [u, flg, rlr, pit] = pcg(@(x) KTWKDTWD(x, Nr, Nc, K, KT, WF, WR, lambda),wfs,pcgtol,pcgitn,[],[],u);
    U = reshape(u,Nr,Nc);

    % Compute and display functional values and CG status
    is = iterdisp(S, K, lambda, p, q, U, WF, WR, [], relres, flg, pit, [], dflag);
    itstat = [itstat; is];

  end

end

% Final part of status display
iterdispclose(dflag);

return



% Vectorise image v
function u = vec(v)

  u = v(:);
  
return


% Consider operator Dy to be the y derivate of a vectorised
% image. Apply this operator to the unvectorised image and return
% a gradient image.
function u = Dy(v)

  u = [diff(v); zeros(1,size(v,2))];
    
return


% Consider operator Dy to be the y derivate of a vectorised
% image. Apply the transpose of this operator to the unvectorised
% image and return a gradient image.
function u = DyT(v)

  u0 = -v(1,:);
  u1 = -diff(v);
  u2 = v(end-1,:);
  u = [u0; u1(1:(end-1),:); u2];
    
return


% Consider operator Dx to be the x derivate of a vectorised
% image. Apply this operator to the unvectorised image and return
% a gradient image.
function u = Dx(v)

  u = [diff(v,1,2) zeros(size(v,1),1)];
    
return


% Consider operator Dx to be the x derivate of a vectorised
% image. Apply the transpose of this operator to the unvectorised
% image and return a gradient image.
function u = DxT(v)

  u = DyT(v')';

return


% Compute scalar function f_F
function y = fF(x, p, epsf)

  absx = abs(x);
  absxp = absx.^(p-2);
  absxp(isnan(absxp)) = 0;
  absxp(isinf(absxp)) = 0;
  y0 = (absx > epsf).*absxp;
  y1 = (absx <= epsf).*(ones(size(absx))*epsf^(p-2));
  y = y0 + y1;

return


% Compute scalar function f_R
function y = fR(x, q, epsr)

  absx = abs(x);
  absxq = absx.^((q-2)/2);
  absxq(isnan(absxq)) = 0;
  absxq(isinf(absxq)) = 0;
  y = (absx > epsr).*absxq;

return


% Compute I + lambda*D_x^T*D_x + lambda*D_y^T*D_y for vectorised
% image (since called by pcg function)
function u = IDTD(v, Nr, Nc, lambda)
  
  V = reshape(v, Nr, Nc);
  U = V + lambda*DxT(Dx(V)) + lambda*DyT(Dy(V));
  u = U(:);

return


% Compute KT*K + lambda*D_x^T*D_x + lambda*D_y^T*D_y for vectorised
% image (since called by pcg function)
function u = KTKDTD(v, Nr, Nc, K, KT, lambda)
 
  V = reshape(v, Nr, Nc);
  U = KT(K(V)) + lambda*DxT(Dx(V)) + lambda*DyT(Dy(V));
  u = U(:);

return


% Compute I + lambda*D_x^T*W_R*D_x + lambda*D_y^T*W_R*D_y for vectorised image
% (since called by pcg function) 
function u = IDTWD(v, Nr, Nc, WF, WR, lambda)
  
  V = reshape(v, Nr, Nc);
  U = WF.*V + lambda*DxT(WR.*Dx(V)) + lambda*DyT(WR.*Dy(V));
  u = U(:);

return


% Compute I + lambda*W_F^(-1/2)*D_x^T*W_R*D_x*W_F^(-1/2) +
% lambda*W_F^(-1/2)*D_y^T*W_R*D_y*W_F^(-1/2) for vectorised image
% (since called by pcg function) 
function u = IWDTWDW(v, Nr, Nc, WF, WR, lambda)
  
  V = reshape(v, Nr, Nc);
  WFN2 = WF.^(-1/2);
  U = V + lambda*WFN2.*DxT(WR.*Dx(WFN2.*V))+lambda*WFN2.*DyT(WR.*Dy(WFN2.*V));
  u = U(:);

return


% Compute KT*WF*K + lambda*D_x^T*W_R*D_x + lambda*D_y^T*W_R*D_y for
% vectorised image (since called by pcg function) 
function u = KTWKDTWD(v, Nr, Nc, K, KT, WF, WR, lambda)
  
  V = reshape(v, Nr, Nc);
  U = KT(WF.*K(V)) + lambda*DxT(WR.*Dx(V)) + lambda*DyT(WR.*Dy(V));
  u = U(:);

return


% Compute relative residual as used by pcg function
function y = pcgrelres(A,b,x)
  y = norm(A(x) - b)/norm(b);
return


% Compute generalised TV functional value
function [fnc, df, reg] = gentvfnctnl(S, K, lambda, p, q, U)

  s = S(:);
  u = U(:);
  dxu = Dx(U);
  dyu = Dy(U);
  if isempty(K),
    df = sum(abs(u-s).^p)/p;
  else
    df = sum(vec(abs(K(U)-S).^p))/p;
  end
  reg = sum(abs(sqrt(vec(dxu.^2 + dyu.^2))).^q)/q;
  fnc = df + lambda*reg;

return
  

% Compute weighted approximation to generalised TV functional value
function [fnc, df, reg] = gentvwtfnctnl(S, K, lambda, p, q, U, WF, WR)
  
  s = S(:);
  u = U(:);
  dxu = Dx(U);
  dyu = Dy(U);
  if isempty(K),
    df = sum(abs(vec(WF.^(1/2)).*(u-s)).^2)/2;
  else
    df = sum(abs(vec(WF.^(1/2)).*vec(K(U)-S)).^2)/2;
  end
  reg = sum(abs([vec((WR.^(1/2)).*dxu); vec((WR.^(1/2)).*dyu)]).^2)/2;
  fnc = df + lambda*reg;
  
return


% Compute weighted approximation to generalised TV functional value
% for denoising with substitution (indirect) method
function [fnc, df, reg] = gentvdnwtfnctnl(S, lambda, p, q, V, WF, WR)
 
  s = S(:);
  v = V(:);
  dxv = (WR.^(1/2)).*Dx((WF.^(-1/2)).*V);
  dyv = (WR.^(1/2)).*Dy((WF.^(-1/2)).*V);
  df = sum(abs(v - (vec(WF.^(1/2)).*s)).^2)/2;
  reg = sum(abs([vec(dxv); vec(dyv)]).^2)/2;
  fnc = df + lambda*reg;

return


% Open iteration status display
function iterdispopen(dspflag)
  if dspflag,
    disp('Itn  Fnc        DFid       Reg        WFnc       WDFid      WReg       RelRes    CGIt  Flg Time');
    disp('--------------------------------------------------------------------------------------------------');
  end
return


% Compute and display iteration status
function is = iterdisp(S, K, lambda, p, q, U, WF, WR, k, relres, pcgflg, pcgit, sbstflg, dspflag)

  t = toc;
  [tnp, dfnp, rnp] = gentvfnctnl(S, K, lambda, p, q, U);
  if isempty(K) && sbstflg,
    V = WF.^(1/2).*U;
    [tnw, dfnw, rnw] = gentvdnwtfnctnl(S, lambda, p, q, V, WF, WR);
  else
    [tnw, dfnw, rnw] = gentvwtfnctnl(S, K, lambda, p, q, U, WF, WR);
  end
  if dspflag,
    kstr = '   ';
    if ~isempty(k),
      kstr = sprintf('%3d', k);
    end
    pistr = '     ';
    if ~isempty(pcgit),
      pistr = sprintf('%5d', pcgit);
    end
    pfstr = '  ';
    if ~isempty(pcgflg),
      pfstr = sprintf('%2d', pcgflg);
    end
    disp(sprintf('%s %10.3e %10.3e %10.3e %10.3e %10.3e %10.3e %10.3e %s %s %8.1e', kstr, tnp, dfnp, rnp, tnw, dfnw, rnw, relres, pistr, pfstr, t));
  end
  is = [tnp dfnp rnp tnw dfnw rnw relres pcgit pcgflg t];

return


% Close iteration status display
function iterdispclose(dspflag)
  if dspflag,
    disp('--------------------------------------------------------------------------------------------------');
  end
return

