function [img, msk, rgn] = inpnttst(imnm, mskn)
    
% inpnttst -- Get an inpainting test image-mask pair
%
% Usage:
%       [img, msk, rgn] = inpnttst(imnm, mskn)
%
% Input:
%       imnm        String containing the image name
%       mskn        String containing the mask id
%
% Output:
%       img         Image
%       msk         Mask
%       rgn         Coordinates of masked region if it is a single
%                   rectangular region, or of a rectangular region
%                   enclosing the masked region if it is a single
%                   contiguous region of limited extent.

if nargin < 2,
  mskn = [];
end

barbr4s1 = @() ellipse([576 720],269,177,6,9);
barbr4s2 = @() ellipse([576 720],269,177,10,20);
barbr4s3 = @() ellipse([576 720],269,177,15,30);

mskc = {
    {'step',    @step,    {'r1s1', @stepmsk, [115 99 144 158]}},...
    {'bars1',   @bars1,   {'r1s1', [120 113 136 140]},...
                          {'r2s1', @bars1msk, [109 109 148 148]}},...
    {'bars2',   @bars2,   {'r1s1', [120 113 136 140]}},...
    {'circle',  @circle,  {'r1s1', [75 75 94 94]}},...
    {'barbara', [],       {'r1s1', [308 560 317 589]},...
                          {'r1s2', [306 560 319 589]},...
                          {'r1s3', [300 550 319 589]},...
                          {'r1s4', [300 550 324 593]},...
                          {'r2s1', [346 354 357 383]},...
                          {'r2s2', [346 354 361 383]},...
                          {'r3s1', [67 527 76 566]},...
                          {'r4s1', barbr4s1, [263 168 275 186]},...
                          {'r4s2', barbr4s2, [259 157 279 197]},...
                          {'r4s3', barbr4s3, [254 147 284 207]},...
                          {'r5s1', [273 282 300 312]},...
                          {'r6s1', [287 454 313 481]}},...
    {'lena',    [],       {'r1s1', [44 245 63 252]},...
	                  {'r1s2', [44 243 63 254]}},...
    {'kodim23', [],       {'r1s1', []},...
                          {'r1s2', []},...
                          {'r1s3', []}}
       };


slm = {};
for k = 1:length(mskc),
  if strcmp(mskc{k}{1}, imnm),
    if isempty(mskn),
      slm = {mskc{k}{1} mskc{k}{2} mskc{k}{3}{:}};
    else
      for l = 3:length(mskc{k}),
	if strcmp(mskc{k}{l}{1}, mskn),
          %slm = {mskc{k}{1} mskc{k}{2} mskc{k}{l}{1} mskc{k}{l}{2}};
	  slm = {mskc{k}{1} mskc{k}{2} mskc{k}{l}{:}};
          break;
	end
      end
    end
    break;
  end
end

if isempty(slm),
  img = []; msk = []; rgn = [];
else
  if isempty(slm{2}),
    img = stdimage(slm{1});
  else
    img = feval(slm{2});
  end
  rgn = slm{4};
  if isa(rgn, 'function_handle'),
    msk = feval(rgn);
    rgn = [];
    if length(slm) == 5,
      rgn = slm{5};
    end
  elseif isempty(rgn),
    msk = stdimage(sprintf('%s.%s',slm{1},slm{3}),1,{'inptmsk'});
    if max(msk(:)) > 1,
      msk = msk / max(msk(:));
    end
    rgn = [];
  else
    msk = ones(size(img,1),size(img,2));
    msk(rgn(1):rgn(3),rgn(2):rgn(4)) = 0;
  end
end

return



function img = step()

  Nr = 256; Nc = 256;
  img = ones(Nr,Nc);
  img(floor(Nr/2):end,:) = 0.5;

return


function msk = stepmsk()

  Nr = 256; Nc = 256;
  Ncr0 = 30; Ncr1 = 35; Ncy1 = 20; Ncy = 16;
  [gx,gy] = meshgrid((1:Nr)-(Nr+1)/2,(1:Nc)-(Nc+1)/2);
  msk1 = ones(Nr,Nc);
  msk1(gx.^2 + (gy-Ncy).^2 <= Ncr0.^2 & (gy-Ncy) < 0) = 0;
  msk2 = ones(Nr,Nc);
  msk2(gx.^2 + (gy-Ncy-Ncy1).^2 <= Ncr1.^2 & (gy-Ncy) < 0) = 0;
  msk = (msk1 + msk2 ~= 1);

return


function img = bars1()

  Nr = 256; Nc = 256; Nbw = 5;
  img = 0.5*ones(Nr,Nc);
  for k = 1:(3*Nbw):(Nc-2*Nbw+1), 
    img(:,(k+Nbw):(k+2*Nbw-1)) = 1; 
  end

return


function msk = bars1msk()

  Nr = 256; Nc = 256; Nmh = 8; Nmw = 20;
  Nr2 = floor(Nr/2); Nc2 = floor(Nc/2);
  msk = ones(Nr,Nc);
  msk((Nr2-Nmh+1):(Nr2+Nmh),(Nc2-Nmw+1):(Nc2+Nmw)) = 0;
  msk = round(imrotate(msk,45,'bilinear','crop'));
  [gx,gy] = meshgrid((1:Nr)-(Nr+1)/2,(1:Nc)-(Nc+1)/2);
  msk(abs(gx) > 2*max(Nmh,Nmw) & abs(gy) > 2*max(Nmh,Nmw)) = 1;

return


function img = bars2()

  Nr = 256; Nc = 256; Nbw = 5; Ncg = 32;
  img = 0.5*ones(Nr,Nc);
  for k = 1:(3*Nbw):(Nc-2*Nbw+1), 
    img(Ncg:(Nc-Ncg),(k+Nbw):(k+2*Nbw-1)) = 1; 
  end

return


function img = circle()
  
  Ncr0 = 64; Ncr1 = 112;
  img = bars1();
  [Nr Nc] = size(img);
  [gx,gy] = meshgrid((1:Nr)-(Nr+1)/2,(1:Nc)-(Nc+1)/2);
  img(gx.^2 + gy.^2 <= Ncr1.^2) = 0.5;
  img(gx.^2 + gy.^2 <= Ncr0.^2) = 1;

return


function msk = ellipse(sz, cy, cx, ry, rx)
  
  [x,y] = meshgrid(1:sz(2),1:sz(1));
  msk = ((x-cx).^2)/rx^2 + ((y-cy).^2)/ry^2 <= 1;
  msk = ~msk;

return
